unit TDropV;
(*
               TreeDropView Component (Object Pascal - Delphi 2.0)
                           (c) Daniel Doubrovkine
  Stolen Technologies Inc. - University of Geneva - 1996 - All Rights Reserved
                       for XReplace-32, version 1.51

             This source code is free for educational purposes only.
             If you intend to use this code in an application other
             than freeware, please contact the author for agreement.

                   e-mail welcome: dblock@infomaniak.ch
                         no warranty of any kind
*)
interface

uses
  Classes, ComCtrls, FileCtrl, Controls, Forms, Dialogs, SysUtils, Buttons, WinTypes, Messages;

type
  {--- the TDropView class, descendant from TTreeView ---}
  TDropView = class(TTreeView)
  private
     {public TreeDropView drive node pointer}
     DropTreeLocal: boolean;
     FStatusBar: TStatusBar;
     FOnDropFinished: TNotifyEvent;
     {the drop aborter}
     InterruptDrop: boolean;
     {local drive node pointer, sort of temporary}
     LocalNode: TTreeNode;
     {dynamic items used not to block the operations on file list boxes}
     NewFilterCombo: TFilterComboBox; {filter combo box}
     NewFileList: TFileListBox;       {file list box}
     {an external TFilterCombo is used in case none is attached, so no special
      code has to be written}
     DisposeFilter: boolean;
     {if this is True, subdirectories are dropped, otherwise only selected
      directory is dropped from DirectoryListBox}
     FDropTree: boolean;
     {if this is True, empty directories are also added, otherwise the are
      deleted as recursion finds no files to add}
     FEmpty: boolean;
     {indicates user can delete an item dropped}
     FDelete: boolean;
     {the filter combo to use, if a filter combo is attached, the files of type
      in it only will be added}
     FFilterCombo: TFilterComboBox;
     {drag drop functions, overriding TTreeView dragdrops with private functions}
     procedure DragDrop(Source: TObject; X, Y: Integer); override;
     procedure PrivateDragDrop(Sender, Source: TObject; X, Y: Integer);
     {drag over functions, overriding TTreeView dragovers with private functions}
     procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
     procedure PrivateDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
     {keyup functions, overriding the TTreeView keyups, for item deletion with Del key}
     procedure KeyUp(var Key: Word; Shift: TShiftState); override;
     procedure PrivateKeyUp(var Key: Word; Shift: TShiftState);
     {setting the newly declared property values}
     procedure SetDropTree(Value: boolean);
     procedure SetEmpty(Value: boolean);
     procedure SetDelete(Value: boolean);
     procedure Status(StatString: string);
     procedure UpdateGlyphs;
  public
     {updated by the adder thread indicated that it has finished the drop operation}
     DropTerminated: boolean;
     constructor Create(AOwner: TComponent); override; {no comment ...}
  published
     {the associated FilterComboBox affects the type of the files droppped}
     property FilterCombo: TFilterComboBox read FFilterCombo write FFilterCombo default nil;
     {check FDropTree}
     property DropTree: boolean read FDropTree write SetDropTree default True;
     {check FEmpty}
     property Empty: boolean read FEmpty write SetEmpty default True;
     {check FDelete}
     property CanDelete: boolean read FDelete write SetDelete default True;
     {kill the current drop operation if any, does not generate an exception
      if no operation running}
     procedure Kill;
     {get the complete path of a node}
     function GetCompletePath(ANode: TTreeNode): string;
     property OnDropFinished: TNotifyEvent read FOnDropFinished write FOnDropFinished;
     property StatusBar: TStatusBar read FStatusBar write FStatusBar;
     procedure LoadFromFile(const FileName: string);
     procedure HalfExpand;
  end;

 {--- the multithread adder to the TDropView ---}
  AdderThread = class(TThread)
  private
     Drive: char;                  {working drive}
     TreeDropView: TDropView;      {dropping target}
     FileListBox: TFileListBox;    {dynamic file list box}
     InitDirectory: string;        {initial drop directory}
     {the dragdrop executer}
     procedure TreeDropDirectory(CurrentDirectory: string);
     {the multithread default executor}
     procedure Execute; override;
     {check the subnode existance function}
     function  ATreeNodeExistanceCheck(var ATreeNode : TTreeNode;Anode : String): Boolean;
  public
     {the multithread drop constructor}
     constructor Create(TrDropView: TDropView; FilListBox: TFileListBox; CurDir: string);
     end;

 procedure Register;                                        {component register}

implementation

{--- TreeDropView ----------------------------------------------------------------------}
procedure TDropView.PrivateDragDrop(Sender, Source: TObject; X, Y: Integer);
var
   i: integer;
begin
   DropTreeLocal:=DropTree;
   if Source is TFileListBox then
   if (Hi(GetKeyState(VK_SHIFT))<>0) then begin
      with (Source as TFileListBox) do LoadFromFile(Items[ItemIndex]);
      FullExpand;
      exit;
      end;
   if DropTree and (Hi(GetKeyState(VK_CONTROL))<>0) then DropTreeLocal:=False;
   {set default state}
   InterruptDrop:=False;
   DropTerminated:=False;
   NewfileList:=nil;
   {create a filter combo box at the image of the assigned one if any}
   NewFilterCombo:=TFilterComboBox.Create(Application);
   with NewFiltercombo do begin
      Parent:=Application.MainForm;
      Visible:=False;
      if not Assigned(FFilterCombo) then
         Filter:='(*.*)|*.*' else
         Filter:=FFilterCombo.Filter;
      end;
   {create a file list box}
    NewFileList:=TFileListBox.Create(Application);
    with NewfileList do begin
       Visible:=False;
       Parent:=Application.MainForm;
       MultiSelect:=True;
       Update;
       end;
   if Source is TDirectoryListBox then begin
     {dropping from DirectoryListBox}
      with (Source as TDirectoryListBox) do begin
         Directory:=GetItemPath(ItemIndex);
         with NewfileList do begin
            Directory:=(Source as TDirectoryListBox).Directory;
            Update;
            end;
         AdderThread.Create(Self, nil, Directory);     {go multithread}
         end;
      end else
   if Source is TFileListBox then begin
     {dropping from FileListBox}
      with (Source as TFileListBox) do begin
         {create a file list box at the image of the source}
         with NewfileList do begin
            FileType:=(Source as TFileListBox).FileType;
            Directory:=(Source as TFileListBox).Directory;
            Mask:=(Source as TFileListBox).Mask;
            Update;
            for i:=0 to Items.Count-1  do Selected[i]:=(Source as TFileListBox).Selected[i];
            end;
            AdderThread.Create(Self, NewfileList, Directory); {go multithread}
            end;
      end;
   {this is really ugly, wait for the drop to terminate}
   while not DropTerminated do Application.ProcessMessages;
   {drop has finished, dynamic stuff can rest in peace}
   NewFilterCombo.Free;
   NewFileList.Free;
   Status('Ready.');
   end;

procedure TDropView.PrivateDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
  if DropTerminated then                    {avoid dropping on a working TDropView}
  if (Source is TFileListBox) or
     (Source is TDirectoryListBox) then
     Accept:=True else Accept:=False;
     end;

procedure TDropView.PrivateKeyUp(var Key: Word; Shift: TShiftState);
begin
   try
   if DropTerminated then                 {avoid deleting an item on a working TDropView}
   if Key=46 {delete} then begin
      Items.Delete(Selected);
      end;
   except
   end;
   end;


procedure TDropView.DragDrop(Source: TObject; X, Y: Integer);
begin
   inherited DragDrop(Source, X, Y);
   PrivateDragDrop(Self, Source, X, Y);
   if Assigned(FOnDropFinished) then FOnDropFinished(Self);
   end;

procedure TDropView.DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
   inherited DragOver(Source, X, Y, State, Accept);
   PrivateDragOver(Self, Source, X, Y, State, Accept);
   end;

procedure TDropView.KeyUp(var Key: Word; Shift: TShiftState);
begin
   inherited KeyUp(Key, Shift);
   if FDelete then PrivateKeyUp(Key, Shift);
   end;

constructor TDropView.Create(AOwner: TComponent);
begin
   DropTree:=True;                               {drop entire trees}
   Empty:=False;                                 {don't drop empty directories}
   CanDelete:=True;                              {can delete items}
   DropTerminated:=True;                         {doing nothing now}
   inherited Create(AOwner);                     {go for it}
   end;

{--- checking a node existance ------------------------------------------------------}
function AdderThread.ATreeNodeExistanceCheck(var ATreeNode : TTreeNode; Anode : String): Boolean;
var
   i:integer;
begin
   try
      ATreeNodeExistanceCheck:=False;
      If ATreeNode.Count=0 Then Exit;
      for i:=ATreeNode.Count-1 DownTo 0 do begin
         If CompareText(ATreeNode.Item[i].Text,ANode)=0 Then begin
            AtreeNodeExistanceCheck:=True;
            ATreeNode:=ATreeNode.Item[i];
            break;
            end;
         end;
   except
      ATreeNodeExistanceCheck:=False;
   end;
   end;

{--- multithread adderthread execution ------------------------------------------------}
procedure AdderThread.Execute;
begin
   try
   TreeDropDirectory(InitDirectory);
   finally
   TreeDropView.DropTerminated:=True;
   TreeDropView.Update;
   end;
   end;

{--- the core of the adderthread ------------------------------------------------------}
procedure AdderThread.TreeDropDirectory(CurrentDirectory: string);
   {create a directory node in the DropView}
   procedure ForceDirectoryCreate(var ATreeNode: TTreeNode);
   var
      ANode:string;
   begin
      ANode:=Copy(CurrentDirectory,1,Length(CurrentDirectory));
      Delete(ANode,1,3);
      while (Length(Anode)>0) do
      begin
         If Pos('\',ANode)=0 then begin
            if (ATreeNodeExistanceCheck(ATreeNode,Anode)=False) then begin
               ATreeNode:=TreeDropView.Items.AddChild(ATreeNode,Anode);
               if Assigned(TreeDropView.StateImages) then ATreeNode.StateIndex:=2;
               end;
            break;
            end;
         if (ATreeNodeExistanceCheck(ATreeNode,Copy(Anode,1,Pos('\',ANode)-1))=False) then begin
            AtreeNode:=TreeDropView.Items.AddChild(ATreeNode,Copy(Anode,1,Pos('\',ANode)-1));
            if Assigned(TreeDropView.StateImages) then ATreeNode.StateIndex:=2;
            end;
         Delete(ANode,1,Pos('\',ANode));
         end;
   end;

var
   ShortDirectory: string;
   ATreeNode,OldATreeNode: TTreeNode;
   DirectoryFindResult: Integer;
   MySearchRec: TSearchRec;
   i,l,m:integer;
   ExistanceCheckFlag:boolean;
   AddedItem : boolean;
   AFile, ATempNode : TTreeNode;
begin
   try
   {this is a stupid recursive directory parser, too long to comment in detail}
   if TreeDropView.InterruptDrop then exit;
   if Application.Terminated then exit;

   {next two lines avoid errors with root directories}
   ShortDirectory:=CurrentDirectory;
   if ShortDirectory[Length(ShortDirectory)]='\' then Delete(Shortdirectory,Length(ShortDirectory),1);
   TreeDropView.Status('Parsing '+CurrentDirectory);

   if FileListBox=nil then
   if TreeDropView.DropTreeLocal=True then begin
   {parse the made directory and add the whole structure recursively}
   DirectoryFindResult:= FindFirst(ShortDirectory+'\*.*',faDirectory+faReadOnly+faHidden,MySearchRec);
   while DirectoryFindResult = 0 do begin
   {accept any directory structure except for the system one}
      if (MySearchRec.Attr and faDirectory > 0) and
         (MySearchRec.Name<>'..') and (MysearchRec.Name<>'.') then
         try {in case an unexpected error occurs, we don't want to raise exceptions}
         if TreeDropView.InterruptDrop then exit;
         if Application.Terminated then exit;
         OldATreeNode:=ATreeNode;
         {recurse the node add for subdirectory}
         TreeDropDirectory(ShortDirectory+'\'+MySearchRec.Name);
         ATreeNode:=OldATreeNode;
         except
         end;

      DirectoryFindResult := FindNext(MySearchRec);         {get next filename}
      end;{while}
      SysUtils.FindClose(MySearchRec);
      end;{if DropTree}

   {add a node as the current directory}
   (*-------------------------------------------------------------------------------*)

   with TreeDropView do begin
      LocalNode:=Items.GetFirstNode;
      while LocalNode<>nil do begin
         if CompareText(Drive+':\',LocalNode.Text)=0 then break;
         LocalNode:=LocalNode.GetNextSibling;
         end;
      end;

   If (TreeDropView.Items.Count=0) or (TreeDropView.LocalNode=nil) then begin {add anyway}
      ATreeNode:=TreeDropView.Items.Add(TreeDropView.Selected,UpperCase(Drive)+':\');
      TreeDropView.LocalNode:=ATreeNode;
      TreeDropView.Status('Added drive node, '+UpperCase(Drive)+':\');
      if Assigned(TreeDropView.StateImages) then ATreeNode.StateIndex:=1;
      end else ATreeNode:=TreeDropView.LocalNode;

   {--- reach the node of the directory, if does not exist, construct it ------------}
   AddedItem:=False;
   ForceDirectoryCreate(ATreeNode);
   (*-------------------------------------------------------------------------------*)
   {construct the file names consequently at this node}
   if FileListBox<>nil then begin {single or multiple select file list drop}
      if ATreeNode.Count=0 Then ExistanceCheckFlag:=False else ExistanceCheckFlag:=True;
      for l:=0 To FileListBox.Items.Count-1 do begin
         if FileListBox.Selected[l] then begin
            if TreeDropView.InterruptDrop then exit;
            if Application.Terminated then exit;
            i:=-1;
            if ExistanceCheckFlag=True then
            for i:=ATreeNode.Count-1 DownTo 0 do
               if CompareText(ATreeNode.Item[i].Text,FileListBox.Items[l])=0 Then break;

            if (i<0) then begin
               AFile:=TreeDropView.Items.AddChild(ATreeNode,FileListBox.Items[l]);
               TreeDropView.Status('Added '+ShortDirectory+'\'+AFile.Text);
               if Assigned(TreeDropView.StateImages) then AFile.StateIndex:=3;
               AddedItem:=True;
               end;
            end;{if Selected}
         end;{for}
      end else begin {case Sender of FileListBox1}
      TreeDropView.NewFileList.Directory:=ShortDirectory+'\';
      for l:=0 to TreeDropView.NewFilterCombo.Items.Count-1 do begin
         if ATreeNode.Count=0 then ExistanceCheckFlag:=False else ExistanceCheckFlag:=True;
         TreeDropView.NewFilterCombo.ItemIndex:=l;
         TreeDropView.NewFileList.Mask:=TreeDropView.NewFilterCombo.Mask;
         for m:=0 to TreeDropView.NewFileList.Items.Count-1 do begin
            {add file names with double checking}
            if TreeDropView.InterruptDrop then exit;
            if Application.Terminated then exit;
            {speed up a little the item add}
            if ExistanceCheckFlag=True then begin
            for i:=ATreeNode.Count-1 downto 0 do begin
               if CompareText(ATreeNode[i].Text,TreeDropView.NewFileList.Items[m])=0 then break;
               end;
            if (i<0) then begin
               AFile:=TreeDropView.Items.AddChild(ATreeNode,TreeDropView.NewFileList.Items[m]);
               TreeDropView.Status('Added '+ShortDirectory+'\'+AFile.Text);
               if Assigned(TreeDropView.StateImages) then AFile.StateIndex:=3;
               AddedItem:=True;
               end;
            end else begin
            AFile:=TreeDropView.Items.AddChild(ATreeNode,TreeDropView.NewFileList.Items[m]);
            TreeDropView.Status('Added '+ShortDirectory+'\'+AFile.Text);
            if Assigned(TreeDropView.StateImages) then AFile.StateIndex:=3;
            AddedItem:=True;
            end;

      if TreeDropView.InterruptDrop then exit;
      if Application.Terminated then exit;
      end;{for}
   end; {for}
   end; {case}

if (not AddedItem) and (not TreeDropView.FEmpty) then begin
   ATempNode:=ATreeNode;
   while ATempNode<>nil do begin
      ATreeNode:=ATempNode.Parent;
      if ATempNode.Count=0 then ATempNode.Delete else break;
      ATempNode:=ATreeNode;
      end;
   end;

except
end;
end;{TreeDropDirectory}

{--- adder thread constructor ---------------------------------------------------------}
constructor AdderThread.Create(TrDropView: TDropView; FilListBox: TFileListBox; CurDir: string);
begin
   FreeOnTerminate:=True;
   FileListBox:=FilListBox;
   TreeDropView:=TrDropView;
   InitDirectory:=CurDir;
   Drive:=InitDirectory[1];
   inherited Create(False);
   end;

procedure TDropView.Kill;
begin
   InterruptDrop:=True;
   end;

procedure TDropView.SetEmpty(Value: boolean);
begin
   if Value<>FEmpty then begin
      FEmpty:=Value;
      Update;
      end;
   end;

procedure TDropView.SetDropTree(Value: boolean);
begin
   if Value<>FDropTree then begin
      FDropTree:=Value;
      Update;
      end;
   end;

procedure TDropView.SetDelete(Value: boolean);
begin
   if Value<>FDelete then begin
      FDelete:=Value;
      Update;
      end;
   end;

function TDropView.GetCompletePath(ANode: TTreeNode): string;
begin
   if ANode.Parent<>nil then begin
      Result:=GetCompletePath(ANode.Parent);
      if Result[Length(Result)]<>'\' then Result:=Result+'\';
      Result:=Result+ANode.Text
   end else Result:=ANode.Text;
   end;

procedure TDropView.Status(StatString: string);
begin
   try
   if Assigned(StatusBar) then begin
      StatusBar.SimpleText:=StatString;
      StatusBar.Update;
      end;
   except
   end;
   end;

procedure TDropView.UpdateGlyphs;
   procedure RecurseNode(ANode: TTreeNode);
   begin
      while ANode<>nil do begin
         if ANode.Count>0 then begin
            RecurseNode(ANode.GetFirstChild);
            if ANode.StateIndex<=0 then ANode.StateIndex:=2;
            end else
         if DirectoryExists(GetCompletePath(ANode)) then ANode.StateIndex:=2 else
            ANode.StateIndex:=3;
         ANode:=ANode.GetNextSibling;
         end;
      end;
var
   ANode: TTreeNode;
begin
   if Assigned(StateImages) then begin
      Status('Updating glyphs...');
      ANode:=Items.GetFirstNode;
      while(ANode<>nil) do begin
         ANode.StateIndex:=1;
         RecurseNode(ANode);
         ANode:=ANode.GetNextSibling;
         end;
      end;
      Status('Ready...');
   end;

procedure TDropView.LoadFromFile(const FileName: string);
var
   ANode: TTreeNode;
   AText: string;
begin
   Status('Loading file...');
   inherited;
   ANode:=Items.GetFirstNode;
   while ANode<>nil do begin
      AText:=ANode.Text;
      Delete(AText,1,1);
      if AText<>':\' then begin
         MessageDlg('ERROR! Attemted to load an invalid file container!',mtError,[mbOk],0);
         Items.Clear;
         break;
         end;
      ANode:=ANode.GetNextSibling;
      end;
   UpdateGlyphs;
   end;


procedure TDropView.HalfExpand;
   function ExpandDirectories(ANode: TTreeNode): boolean;
   begin
      Result:=True;
      while ANode<>nil do begin
         if ANode.Count>0 then begin
            if ExpandDirectories(ANode.GetfirstChild) then
               try
               Anode.Expand(False)
               except
               end
            else
               try
               ANode.Collapse(True);
               except
               end
            end else
            if not DirectoryExists(GetCompletePath(ANode)) then Result:=False;
         ANode:=ANode.GetNextSibling;
      end;
      end;
begin
   Status('Half expanding directores...');
   ExpandDirectories(Items.GetFirstNode);
   Status('Ready.');
   end;

{--------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Samples', [TDropView]);
end;

end.
